/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

import java.io.*;

/**
 * Read a Standard MIDI File.
 * There are a couple static convenience methods
 * for reading a MIDI file with a single statement, e.g.
 * <pre>MidiPhrase p = MidiFileReader.readMidiPhrase("jsbach.mid");</pre>
 * <p>If you want to get more control, create a MidiFileReader
 * object, use setAction() to set its MidiFileAction object,
 * and call read().
 */

public class MidiFileReader {

	boolean nomerge = false;	// true means continue'ed system
					//  exclusives are not collapsed.
	long currtime = 0;		// current time in delta-time units

	boolean skipinit = true;	// 1 if initial garbage should be skipped

	int toberead = 0;
	int Mformat;
	MidiFileAction action;
	InputStream file;

	private static final int EOF = -1;

	public MidiFileReader(InputStream in) {
		file = in;
	}

	public void setAction(MidiFileAction a) {
		action = a;
	}

	/**
	 * Read the entire MIDI File.
	 * @exception	MidiException	If there is an error while reading.
	 */
	public void read() throws MidiException {
		int ntrks = readheader();
		if ( ntrks <= 0 )
			mfwarning("No tracks!");
		while ( ntrks-- > 0 ) {
			if ( readtrack() == EOF )
				break;
		}
	}

	private void mfwarning(String s) {
		System.err.println(s);
	}

	int mgetc() throws MidiException {
		try {
			int c = file.read();
			return c;
		} catch (IOException io) {
			throw new MidiException("IOException while reading MidiFile!? "+io.getMessage());
		}
	}

	// read a single character and abort on EOF
	private int egetc() throws MidiException {
		int c = mgetc();
		if ( c == EOF ) {
			throw new MidiException("premature EOF");
		}
		toberead--;
		return(c);
	}

	// readvarinum - read a varying-length number, and return the
	// number of characters it took.

	private long readvarinum() throws MidiException {
		long value;
		int c;

		c = egetc();
		value = c;
		if ( (c & 0x80) != 0 ) {
			value &= 0x7f;
			while ( true ) {
				c = egetc();
				value = (value << 7) + (c & 0x7f);
				if ( (c & 0x80) == 0 )
					break;
			}
		}
		return (value);
	}

	private int to32bit(int c1,int c2,int c3,int c4) {
		int value = 0;
	
		value = (c1 & 0xff);
		value = (value<<8) + (c2 & 0xff);
		value = (value<<8) + (c3 & 0xff);
		value = (value<<8) + (c4 & 0xff);
		return (value);
	}
	
	private int to16bit(int c1,int c2) {
		return ((c1 & 0xff ) << 8) + (c2 & 0xff);
	}
	
	private int read32bit() throws MidiException {
		int c1, c2, c3, c4;
	
		c1 = egetc();
		c2 = egetc();
		c3 = egetc();
		c4 = egetc();
		return to32bit(c1,c2,c3,c4);
	}
	
	private int read16bit() throws MidiException {
		int c1, c2;
		c1 = egetc();
		c2 = egetc();
		return to16bit(c1,c2);
	}

	private byte Msgbuff[];	/* message buffer */
	
	private void msginit() {
		Msgbuff = new byte[0];
	}

	private byte[] msg() {
		return(Msgbuff);
	}
	
	private void
	msgadd(int c)
	{
		// This code should be changed to allocate space in chunks,
		// rather than allocating a new array for each new byte.

		int oldleng = Msgbuff.length;
		byte newmess[] = new byte[oldleng+1];
		/* copy old message into larger new one */
		System.arraycopy(Msgbuff,0,newmess,0,oldleng);
		Msgbuff = newmess;

		Msgbuff[oldleng] = (byte)c;
	}

	/* read through the "MThd" or "MTrk" header string */
	/* if skip is true, we attempt to skip initial garbage. */
	private int readmt(String s,boolean skip) throws MidiException {

		int nread = 0;
		char[] b;
		int c;
	
		b = new char[4];
	    retry:
		while ( true ) {
			while ( nread<4 ) {
				c = mgetc();
				if ( c == EOF ) {
					mfwarning("EOF while expecting "+s);
					return(EOF);
				}
				b[nread++] = (char)c;
			}
			/* See if we found the 4 characters we're looking for */
			if ( s.charAt(0)==b[0] && s.charAt(1)==b[1] && s.charAt(2)==b[2] && s.charAt(3)==b[3] )
				break;
			if ( skip == false )
				throw new MidiException("expecting "+s);
			/* If we are supposed to skip initial garbage, */
			/* try again with the next character. */
			b[0]=b[1];
			b[1]=b[2];
			b[2]=b[3];
			nread = 3;
		}
		return(0);
	}
	
	/* read a header chunk */
	private int readheader() throws MidiException {

		int format, ntrks, division;
	
		if ( readmt("MThd",skipinit) == EOF )
			return(0);
	
		toberead = read32bit();
		format = read16bit();
		ntrks = read16bit();
		division = read16bit();
	
		action.k_header(format,ntrks,division);
	
		/* flush any extra stuff, in case the length of header is not 6 */
		while ( toberead > 0 )
			egetc();
		return(ntrks);
	}
	
	private void metaevent(int type) {

		byte[] m = msg();
	
		switch  ( type ) {
		case 0x00:
			action.k_seqnum(to16bit((int)(m[0]),(int)(m[1])));
			break;
		case 0x01:	/* Text event */
		case 0x02:	/* Copyright notice */
		case 0x03:	/* Sequence/Track name */
		case 0x04:	/* Instrument name */
		case 0x05:	/* Lyric */
		case 0x06:	/* Marker */
		case 0x07:	/* Cue point */
		case 0x08:
		case 0x09:
		case 0x0a:
		case 0x0b:
		case 0x0c:
		case 0x0d:
		case 0x0e:
		case 0x0f:
			/* These are all text events */
			action.k_metatext(type,m);
			break;
		case 0x20:	/* Channel prefix */
		case 0x21:	/* Supposedly some people mistakenly used 0x21 ? */
			action.k_chanprefix(m[0]);
			break;
		case 0x2f:	/* End of Track */
			/* action.k_eot(); */
			break;
		case 0x51:	/* Set tempo */
			action.k_tempo(to32bit(0,(int)m[0],(int)m[1],(int)m[2]));
			break;
		case 0x54:
			action.k_smpte(m[0],m[1],m[2],m[3],m[4]);
			break;
		case 0x58:
			action.k_timesig(m[0],m[1],m[2],m[3]);
			break;
		case 0x59:
			action.k_keysig(m[0],m[1]);
			break;
		case 0x7f:
			/* action.k_sqspecific(m); */
			break;
		default:
			/* action.k_metamisc(type,m); */
			break;
		}
	}
	
	private void chanmessage(int status,int c1,int c2) {

		int chan = 1 + (status & 0xf);
	
		switch ( status & 0xf0 ) {
		case 0x80:
			action.k_noteoff(chan,c1,c2);
			break;
		case 0x90:
			action.k_noteon(chan,c1,c2);
			break;
		case 0xa0:
			action.k_pressure(chan,c1,c2);
			break;
		case 0xb0:
			action.k_controller(chan,c1,c2);
			break;
		case 0xe0:
			action.k_pitchbend(chan,c1,c2);
			break;
		case 0xc0:
			action.k_program(chan,c1);
			break;
		case 0xd0:
			action.k_chanpressure(chan,c1);
			break;
		}
	}
	
	long currTime() {
		return currtime;
	}

	// This array is indexed by the high half of a status byte.  It's
	// value is either the number of bytes needed (1 or 2) for a channel
	// message, or 0 (meaning it's not  a channel message).
	static int chantype[] = {
		0, 0, 0, 0, 0, 0, 0, 0,		// 0x00 through 0x70
		2, 2, 2, 2, 1, 1, 2, 0		// 0x80 through 0xf0
	};

	// read a track chunk
	private int readtrack() throws MidiException {

		long lookfor, lng;
		int c, c1, type;
		boolean sysexcontinue = false;	// true if last message was an unfinished sysex
		boolean running = false;		// true when running status used
		int status = 0;			// (possibly running) status byte
		int needed;
	
		if ( readmt("MTrk",false) == EOF )
			return EOF;
	
		toberead = read32bit();
		currtime = 0;
	
		action.k_starttrack();
	
		while ( toberead > 0 ) {
	
			currtime += readvarinum();	// delta time
	
			c = egetc();
	
			if ( sysexcontinue==true && c != 0xf7 )
				mfwarning("didn't find expected continuation of a sysex");
	
			if ( (c & 0x80) == 0 ) {	 // running status?
				if ( status == 0 )
					mfwarning("unexpected running status");
				running = true;
			}
			else {
				status = c;
				running = false;
			}
	
			needed = chantype[ (status>>4) & 0xf ];
	
			if ( needed != 0 ) {		// ie. is it a channel message?
	
				if ( running )
					c1 = c;
				else
					c1 = egetc() & 0x7f;
	
				// The &0xf7 here may seem unnecessary, but I've seen
				// 'bad' midi files that had, e.g., volume bytes
				// with the upper bit set.  This code should not harm
				// proper data.
	
				chanmessage( status, c1, (needed>1) ? (egetc()&0x7f) : 0 );
				continue;
			}
	
			switch ( c ) {
	
			case 0xff:			// meta event
	
				type = egetc();
				// watch out - Don't combine the next 2 statements
				lng = readvarinum();
				lookfor = toberead - lng;
				msginit();
	
				while ( toberead > lookfor )
					msgadd(egetc());
	
				metaevent(type);
				break;
	
			case 0xf0:		// start of system exclusive

				// watch out - Don't combine the next 2 statements
				lng = readvarinum();
				lookfor = toberead - lng;
				msginit();
				msgadd(0xf0);
	
				while ( toberead > lookfor )
					msgadd(c=egetc());
	
				if ( c==0xf7 || nomerge==false )
					action.k_sysex(msg());
				else
					sysexcontinue = true;  // merge into next msg
				break;
	
			case 0xf7:	// sysex continuation or arbitrary stuff
	
				// watch out - Don't combine the next 2 statements
				lng = readvarinum();
				lookfor = toberead - lng;
	
				if ( sysexcontinue == false )
					msginit();
	
				while ( toberead > lookfor )
					msgadd(c=egetc());
	
				if ( sysexcontinue == false ) {
					action.k_arbitrary(msg());
				}
				else if ( c == 0xf7 ) {
					action.k_sysex(msg());
					sysexcontinue = false;
				}
				break;
			default:
				mfwarning("unexpected byte: "+Integer.toString(c));
				break;
			}
		}
		action.k_endtrack();
		return 0;
	}

	/**
	 * Read the named MIDI file and return an array of phrases
	 * that contain its tracks.
	 * @param	fname	Name of file to read
	 * @return	MidiPhrase
	 * @exception	MidiException	If there is an error while reading.
	 */
	static public MidiPhrase[] readMidiPhraseArray(String fname) throws MidiException, FileNotFoundException {
		FileInputStream in = new FileInputStream(fname);
		MidiFileReader mfr = new MidiFileReader(in);

		MidiFilePhraseAction mfact = new MidiFilePhraseAction(mfr);
		mfr.setAction(mfact);
		mfr.read();
		MidiPhrase[] arr = mfact.getArray();
		// Set the length of all the phrases to the latest note
		for ( int i=0; i<arr.length; i++ )
			arr[i].setLength(Alg.latest(arr[i]));
		try {
			in.close();
		} catch (Exception e) {
			// ignore
		}
		return arr;
	}

	/**
	 * Read the named MIDI file and return a single phrase (merging
	 * all of its tracks).
	 * @param	fname	Name of file to read
	 * @return	MidiPhrase
	 * @exception	MidiException	If there is an error while reading.
	 */
	static public MidiPhrase readMidiPhrase(String fname) throws MidiException, FileNotFoundException {
		MidiPhrase[] arr = readMidiPhraseArray(fname);
		MidiPhrase p = new MidiPhrase();
		for ( int n=0; n<arr.length; n++ )
			p.merge(arr[n]);
		return p;
	}

	public static void main(String[] args) {
		String fname = args[0];
		MidiPhrase[] arr = null;
		try {
			arr = MidiFileReader.readMidiPhraseArray(fname);
		} catch (Exception e) {
		 	e.printStackTrace();
			System.out.println("Exception, msg="+e.getMessage());
			System.exit(1);
		}
		for ( int n=0; n<arr.length; n++ ) {
			MidiPhrase p = arr[n];
			System.out.println("track n="+n+" p="+p);
		}
	}
}

class MidiNoteUnfinished extends MidiNote {

	final static int NOTHING = 0;
	final static int ON = 1;
	final static int OFF = 2;
	int type;

	MidiNoteUnfinished(int p, long t, int c, int v, long d) {
		super(p,t,c,v,d);
		type = NOTHING;
	}
	void setType(int t) {
		type = t;
	}
	int getType() {
		return type;
	}
	public String toString() {
		return super.toString()+" UN "+Integer.toString(type);
	}
	public Object clone() {
		MidiNoteUnfinished u = new MidiNoteUnfinished(
			this.getPitch(),
			this.getTime(),
			this.getChannel(),
			this.getVelocity(),
			this.getDuration());
		u.type = type;
		return u;
	}

}

class MidiFilePhraseAction implements MidiFileAction {

	MidiPhrase[] phraseArray;
	MidiPhrase Noteq;
	MidiPhrase Currph;
	int Mformat;
	int Numq = 0;
	MidiFileReader mf;
	int Defrelease = 0;
	boolean Onoffmerge = true;
	int Tracknum;
	double Clickfactor = 1.0;
	int Clicks = 96;
	
	MidiFilePhraseAction(MidiFileReader f) {
		mf = f;
	}

	MidiPhrase[] getArray() {
		return phraseArray;
	}

	private long mfclicks() {
		double clks = (double)mf.currTime()/Clickfactor;
		return((long)(clks+0.5)); /* round it */
	}

	int getFormat() {
		return Mformat;
	}

	public void k_header(int f,int n,int d) {
		Mformat = f;
		Tracknum = 0;
		Clickfactor = (double)d / (double)(Clicks);
		phraseArray = new MidiPhrase[n];
	}
	
	public void k_starttrack() {
		// Start a new phrase for the new track
		Currph = new MidiPhrase();
		Noteq = new MidiPhrase();
	}
	
	// output the top Noteq and remove it from the list
	void putnfree() {
		MidiMsg n = Noteq.removeFirstMsg();
		if ( n instanceof MidiNoteUnfinished ) {
			// Convert unfinished ones to normal ones
			if ( ((MidiNoteUnfinished)n).getType() == MidiNoteUnfinished.ON ) {
				n = new MidiNoteOn(n.getPitch(),n.getTime(),n.getChannel(),
								n.getVelocity());
			}
			else {
				n = new MidiNoteOff(n.getPitch(),n.getTime(),n.getChannel(),
								n.getVelocity());
			}
		}
		Currph.merge(n);
		Numq--;
	}

	void putallnotes() {

		while ( Noteq.isEmpty() == false )
			putnfree();
		Currph.setLength(mfclicks());
	}
	
	public void k_endtrack() {
		putallnotes();
		phraseArray[Tracknum++] = Currph;
	}
	
	public void k_noteon(int chan,int pitch,int vol) {

		if ( vol == 0 ) {
			k_noteoff(chan,pitch,Defrelease);
			return;
		}
		queuenote(chan,pitch,vol,0,true);
	}
	
	MidiMsg queuenote(int chan,int pitch,int vol,long dur,boolean on) {

		MidiNoteUnfinished n = new MidiNoteUnfinished(pitch,mfclicks(),chan,vol,dur);

		if ( on == true ) {
			n.setType(MidiNoteUnfinished.ON);
		}
		else
			n.setType(MidiNoteUnfinished.OFF);
		Noteq.merge(n);

		Numq++;
		return n;
	}
	
	public void k_noteoff(int chan,int pitch,int vol) {

		MidiPhrase toputout = new MidiPhrase();
		MidiMsg n = null;
	
		/* find the first note-on (if any) that matches this one */

		MidiPhraseEnumeration e = new MidiPhraseEnumeration(Noteq);
		while ( e.hasMoreElements() ) {
			MidiMsg n1 = (MidiMsg)e.nextElement();
			if ( n1.getChannel()==chan
				&& n1.getPitch()==pitch
				&& n1 instanceof MidiNoteUnfinished
				&& ((MidiNoteUnfinished)n1).getType() == MidiNoteUnfinished.ON ) {
				n = n1;
				break;
			}
		}
		if ( n == null ) {
			/* it's an isolated note-off */
			n = queuenote(chan,pitch,vol,0,false);
		}
		else if ( Onoffmerge == false && vol != Defrelease ) {
			/* If the note-off matches a previous note-on, but has a */
			/* non-default velocity, then we have to turn it into a */
			/* separate keynote note-off, instead of merging it with */
			/* the note-on into a single note. */
			MidiMsg o = queuenote(chan,pitch,vol,0,false);
			n.setDuration(0);
		}
		else {
			// A completed note.
			MidiNote cn = new MidiNote(n.getPitch(),n.getTime(),n.getChannel(),
								n.getVelocity(),mfclicks()-n.getTime());
			// replace the existing note with this one
			Noteq.replace(n,cn);
		}
	
		/* Now start at the beginning of the list and put out any */
		/* notes we've completed.  This guarantees that the starting */
		/* times of the notes are in the proper (ie. monotonically */
		/* progressing) order. */
	
		while ( Noteq.isEmpty() == false ) {
			n = Noteq.getAt(0);
			/* quit when we get to the first unfinished note */
			if ( n instanceof MidiNoteUnfinished )
				break;
			putnfree();
		}
		/* If the number of notes int Noteq gets too big, then we're */
		/* probably suffering from a note-on that never had a note-off.*/
		/* Force it out. */
		if ( Numq > 256 )
			putnfree();
	
	}
	
	public void k_pressure(int chan,int pitch,int press) {
		threebytes(0xa0 | chan,pitch,press);
	}
	
	public void k_controller(int chan,int control,int value) {
		threebytes(0xb0 | chan,control,value);
	}
	
	public void k_pitchbend(int chan,int msb,int lsb) {
		threebytes(0xe0 | chan,msb,lsb);
	}
	
	public void k_program(int chan,int program) {
		twobytes(0xc0 | chan,program);
	}
	
	public void k_chanpressure(int chan,int press) {
		twobytes(0xd0 | chan,press);
	}
	
	void
	threebytes(int c1,int c2,int c3)
	{
		MidiBytes n = new MidiBytes((byte)c1,(byte)c2,(byte)c3);
		n.setTime(mfclicks());
		Noteq.merge(n);
		Numq++;
	}
	
	void
	twobytes(int c1,int c2)
	{
		MidiBytes n = new MidiBytes((byte)c1,(byte)c2);
		n.setTime(mfclicks());
		Noteq.merge(n);
		Numq++;
	}
	
	void
	queuemess(byte[] bytes)
	{
		MidiBytes n = new MidiBytes(mfclicks(),bytes);
		Noteq.merge(n);
		Numq++;
	}
	
	public void k_sysex(byte[] bytes) {
		queuemess(bytes);
	}
	
	public void k_arbitrary(byte[] bytes) {
		queuemess(bytes);
	}
	
	public void k_tempo(int tempo) {
		String s = "Tempo="+Integer.toString(tempo);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	public void k_timesig(int nn,int dd,int cc,int bb) {
		int denom = 1;
	
		while ( dd-- > 0 )
			denom *= 2;
		/* First 2 numbers are time signature, next is MIDI-clocks-per-click, */
		/* and the last is 32nd-notes-per-24-MIDI-clocks. */
		String s = "Timesig="
			+Integer.toString(nn)+","
			+Integer.toString(denom)+","
			+Integer.toString(cc)+","
			+Integer.toString(bb);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	public void k_keysig(int sf,int mi) {
		String s = "Keysig="+Integer.toString(sf)+","+Integer.toString(mi);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	public void k_chanprefix(int c) {
		String s = "Channelprefix="+Integer.toString(c);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	public void k_seqnum(int n) {
		String s = "Sequence="+Integer.toString(n);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	public void k_smpte(int hr,int mn,int se,int fr,int ff) {
		String s = "Smpte="
			+Integer.toString(hr)+","
			+Integer.toString(mn)+","
			+Integer.toString(se)+","
			+Integer.toString(fr)+","
			+Integer.toString(ff);
		Currph.merge(new MidiText(mfclicks(),s));
	}
	
	static String ttypetext[] = {
		"",
		"Text Event",			// type=0x01
		"Copyright Notice",		// type=0x02
		"Sequence/Track Name",	// type=0x03
		"Instrument Name",		// type=0x04
		"Lyric",				// type=0x05
		"Marker",				// type=0x06
		"Cue Point",			// type=0x07
		"Unrecognized"			// type=0x08-0x0f
	};

	public void k_metatext(int type,byte[] mess)
	{
		int unrecognized = ttypetext.length - 1;
	
		if ( type < 1 || type > unrecognized )
			type = unrecognized;
		String s = ttypetext[type];
		for ( int i=0; i<mess.length; i++ ) {
			s += String.valueOf((char)(mess[i]));
		}
		Currph.merge(new MidiText(mfclicks(),s));
	}
}
