/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * Handle MIDI input.
 *
 * @author	Tim Thompson
 */

public class Win32MidiInput implements MidiInput,Runnable {

	int deviceId;
	int deviceHandle = 0;
	MidiFsm fsm;
	boolean merge;
	boolean realtime;
	Thread t;

	/**
	 * Constructor for using input device 0 (default)
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiInput() throws MidiException {
		this(0);
	}

	/**
	 * Constructor for using a particular input device.
	 * @param	id	Device id.
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiInput(int id) throws MidiException {
		// default is for merging and realtime priority
		this(id,true,true);
	}

	/**
	 * Constructor for using a particular input device.
	 * @param	id	Device id.
	 * @param	r	If true, thread uses Windows realtime priority.
	 * @param	m	If true, MIDI input is merged to MIDI output.
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiInput(int id, boolean r, boolean m) throws MidiException {
		deviceId = id;
		realtime = r;
		merge = m;
	}

	/**
	 * Set the object that is invoked when MIDI messages are recognized.
	 * @param	a	The MidiAction object to use.
	 */
	public void setAction(MidiAction a) {
		fsm = new MidiFsm(a);
	}

	/**
	 * Open the device.
	 * @exception MidiException If anything goes wrong.
	 */
	public void open() throws MidiException {

		// only try to open if it is not already open	
 		if (deviceHandle != 0) {
			return;
		}

		// if an attempt to open more than then max number of devices
		// return bad ID error
		if ( deviceId >= MidiSystem.getNumInDevices() ) {
			throw new MidiException("Bad input deviceId - "
				+Integer.toString(deviceId));
		}
    
		// attempt to open the device - it'll throw MidiException if it fails.
		MidiInOpen();
	}

	/**
	 * Reset the MIDI out device.
	 */
	public void reset() {
		try {
		   	MidiInReset();
		} catch (Exception e) {
			System.err.println("MidiInReset error - "+e.getMessage());
		}
	}

	/**
	 * Close the midi device.
	 * @exception MidiException If anything goes wrong.
	 */
	public synchronized void close() throws MidiException {

		if (deviceHandle == 0) {	// not open; nothing to do
			return;
		}
		reset();
		MidiInClose();
	}

	/**
	 * Start the thread that grabs MIDI input.
	 */
	public void startRunning() {
		t = new Thread(this);
		t.start();
	}

	public void run() {
		try {
			MidiInLoop(realtime);
		} catch (Exception e) {
			System.err.println("MidiInLoop error!? - "+e.getMessage());
		}
	}

	/**
	 * Determine whether MIDI input is automatically merged with output.
	 * @param	b	True if MIDI input should be merged with output.
	 */
	public void setMerge(boolean b) {
		merge = b;
	}

	/**
	 * Stop the input thread.
	 */
	public void stopRunning() {
		try {
			MidiInLoopStop();
		} catch (Exception e) {
			System.err.println("MidiInLoopStop error!? - "+e.getMessage());
		}
		t.stop();
	}

	/**
	 * This is called by the MidiIinLoop() native method,
	 * once for each received MIDI input byte.
	 * @param	b	The byte received.
	 */
	public void gotByte(int b) {
		if ( merge ) {
			byte[] bytes = new byte[1];
			bytes[0] = (byte)b;
			try {
				MidiSystem.schedule(bytes);
			} catch (Exception e) { }
		}
		fsm.parse(b);
	}

	/**
	 * Native Methods
	 */
	
	/**
	 * Open the device; if successful, deviceHandle has a valid handle
	 * @returns Device handle.
	 * @exception MidiException If anything goes wrong.
	 */
	private native int MidiInOpen() throws MidiException;

	/**
	 * Reset the device to its default state
	 * @exception MidiException If anything goes wrong.
	 */
	private native int MidiInReset() throws MidiException;

	/**
	 * Close the device
	 * @exception MidiException If anything goes wrong.
	 */
	private native int MidiInClose() throws MidiException;

	/**
	 * Start processing midi input.  This routine does not return
	 * until MidiInLoopStop() is called.
	 * @exception MidiException If anything goes wrong.
	 */
	private native void MidiInLoop(boolean r) throws MidiException;

	/**
	 * Stop processing midi input.
	 * @exception MidiException If anything goes wrong.
	 */
	private native void MidiInLoopStop() throws MidiException;

}
